#!/bin/bash


#########################################################################
#                                                                       #
# ESX/vSphere backup Script, written by Andreas Jutzy                  #
#                                                                       #
# http://www.aj.de                                                      #
#                                                                       #
# USE WITH CARE AND AT YOUR OWN RISK. CHECK INSTRUCTIONS ON WEBSITE     #
#                                                                       #
# Feel free to modify and contribute good ideas!                        #
#                                                                       #
# Contact info@aj.de for questions and to submit new ideas.             #
#########################################################################



#GLOBAL DEBUG VARIABLE
#IF SET TO 1 THEN _NO_ DELETE/BACKUP OPERATIONS ARE PERFORMED
#AND DEBUG MESSAGES WILL BE PRINT INSTEAD

DEBUG=0

#Version of this script
VERSION=2.31

#set global date
DATE=`date +%F`

#set passless keyfile for ssh
IDFILE=/root/.ssh/id_dsa

#where are all your Datastores? defaults to /vmfs/volumes
PATH2LUNS="/vmfs/volumes"

#set ESXs
ESX="esx1 esx2 esx3"

#what is the identifier for your backup Datastores? defaults to BACK
#e.g. LUN0_BACK_PRODUCTION
#or
#e.g. LUN7_SQLMACHINES_BACK
#!!!careful: each Datastore containing the BACKLUNID will be used as a backup Datastore!!!
#!!!THIS MEANS GRACE PERIODS APPLY TO THIS Datastore!!!
BACKLUNID="BACK"

#set backup methods
#ONCE:          make one backup into the ONCE directory. if backup exists, skip backup
#DAILY:         make daily backups into the DAILY directory.
#WEEKLY:        make weekly backups (1st day of week) into the WEEKLY directory
#MONTHLY:       make monthly backups (1st day of month) into the MONTHLY directory
METHODS="ONCE DAILY WEEKLY MONTHLY"

#set keep times
#7 days
KEEPDAILY=168
#7 weeks
KEEPWEEKLY=1176
#7 months
KEEPMONTHLY=4704

#fill variable with ALL Datastores
ALLLUNS="`ls -al $PATH2LUNS|grep "\->"|awk '{print $9,$11}'`"

if [[ -z $ALLLUNS ]]
then
        echo "ERROR: Could not find ANY Datastore. Aborting..."
        exit 1
fi

#declare arrays and functions
declare -a PRODLUNS
declare -a BACKLUNS

#function to get the production Datastores
function get_prodluns ()
{
        loop=0
        while read lunname lunid
        do
                #bash2 workaround for missing =~ (used in esx 3.5)
                if [[ `echo $lunname|grep "$BACKLUNID"` ]]
                then
                        if [[ $DEBUG == 1 ]]
                        then
                                echo "Backup Datastore $lunname found, skipping..."
                        fi
                else
                        if [[ $DEBUG == 1 ]]
                        then
                                echo "Production Datastore $loop $lunname found..."
                        fi
                        #first quick sanity check of Datastore
                        if [[ -d $PATH2LUNS/$lunname ]]
                        then
                                eval "$lunname"=\'"$lunid"\'
                                PRODLUNS[$loop]="$lunname"
                                let loop=$loop+1
                        else
                                echo "ERROR: Found PRODUCTION Datastore $PATH2LUNS/$lunname is not sane. Aborting..."
                                exit 1
                        fi
                fi
        done <<< "$ALLLUNS"
        if  [[ -z ${PRODLUNS[*]} ]]
        then
                echo "ERROR: Could not find a PRODUCTION Datastore. Aborting..."
                exit 1
        else
                echo -e "Found these production Datatsoress ($loop): ${PRODLUNS[*]}\n"
        fi
}

#function to get the backup Datastores using the above declared identifier
function get_backluns
{
        loop=0
        while read lunname lunid
        do
                if [[ `echo $lunname|grep "$BACKLUNID"` ]]
                then
                        if [[ $DEBUG == 1 ]]
                        then
                                echo "Backup Datastore $loop $lunname found..."
                        fi
                        #first quick sanity check of Datastore
                        if [[ -d $PATH2LUNS/$lunname ]]
                        then
                                eval "$lunname"=\'"$lunid"\'
                                BACKLUNS[$loop]="$lunname"
                                let loop=$loop+1
                        else
                                echo "ERROR: Found BACKUP Datastore $PATH2LUNS/$lunname is not sane. Aborting..."
                                exit 1
                        fi

                else
                        if [[ $DEBUG == 1 ]]
                        then
                                echo "Production Datastore $lunname found, skipping..."
                        fi
                fi
        done <<< "$ALLLUNS"
        if  [[ -z ${BACKLUNS[*]} ]]
        then
                echo "ERROR: Could not find a BACKUP Datastore. Aborting..."
                exit 1
        else
                echo -e "Found these backup Datastores ($loop): ${BACKLUNS[*]}\n"
        fi

}

#function to verify BACKUPLUNS
#I implemented that to prevent production datactores to be wiped if something
#went wrong during the detection of datastores
#if a backup datastore does not contain the correct directories, abort
function verify_backluns ()
{
        for i in ${BACKLUNS[*]}
        do
                if [[ ! -d $PATH2LUNS/$i/ONCE ]] || [[ ! -d $PATH2LUNS/$i/DAILY ]] || [[ ! -d $PATH2LUNS/$i/WEEKLY ]] || [[ ! -d $PATH2LUNS/$i/MONTHLY ]]
                then
                        echo "ERROR: BACKUP Datastore $PATH2LUNS/$i is not correctly prepared or NO BACKUP Datastore. Aborting..."
                        exit 1
                else
                        if [[ $DEBUG = 1 ]]
                        then
                                echo "INFO: BACKUP Datatore $PATH2LUNS/$i is correctly prepared."
                        fi
                fi
        done
}

#function to do the real backup
function do_backup ()
{
        case $6 in
        ONCE)
                #find out if this machine has already been backed up to ANY Datastore!
                for i in ${BACKLUNS[*]}
                do
                        GOTCHA=`find $PATH2LUNS/$i/$6/ -name "*$4*"`
                        if [[ $GOTCHA != "" ]]
                        then
                                GOTCHA=`echo $GOTCHA|cut -d ' ' -f 1`
                                break
                        fi
                done

                if [[ $GOTCHA = "" ]]
                then
                        if [[ $DEBUG = 1 ]]
                        then
                                echo -e "I would do now: vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0"
                        else
                                vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0
                        fi
                        echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 to $PATH2LUNS/$2/$6/$3_$4 DONE\n"
                else
                        #A backup has been found
                        echo -e "Backup $6 for $4 with shown name $5 and\nuuid of $1 already exists as $GOTCHA. DONE\n"
                fi
                ;;
        DAILY)
                if [[ $DEBUG = 1 ]]
                then
                        echo -e "I would do now: vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0"
                else
                        vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0
                fi
                echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 to $PATH2LUNS/$2/$6/$3_$4 DONE\n"
                ;;
        WEEKLY)
                #find out if this machine has already been backed up to ANY Datastore!
                for i in ${BACKLUNS[*]}
                do
                        GOTCHA=`find $PATH2LUNS/$i/$6/ -name "*$4*"`
                        if [[ $GOTCHA != "" ]]
                        then
                                GOTCHA=`echo $GOTCHA|cut -d ' ' -f 1`
                                break
                        fi
                done

                if [[ $GOTCHA = "" ]]
                then
                        #no backup found at all, means: make initial backup! fake weekday ;)
                        echo -e "No Backup found for $4 with shown name $5, create initial Backup..."
                        DAYWEEK=1
                fi

                #if it is the 1st day of the week, then backup
                if [[ $DAYWEEK = "1" ]]
                then
                        if [[ $DEBUG = 1 ]]
                        then
                                echo -e "I would do now: vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0"
                        else
                                vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0
                        fi

                        echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 to $PATH2LUNS/$2/$6/$3_$4 DONE\n"
                else
                        echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 skipped, as it is day $DAYWEEK of the week. DONE\n"
                fi
                ;;
        MONTHLY)
                #find out if this machine has already been backed up to ANY Datastore!
                for i in ${BACKLUNS[*]}
                do
                        GOTCHA=`find $PATH2LUNS/$i/$6/ -name "*$4*"`
                        if [[ $GOTCHA != "" ]]
                        then
                                GOTCHA=`echo $GOTCHA|cut -d ' ' -f 1`
                                break
                        fi
                done

                if [[ $GOTCHA = "" ]]
                then
                        #no backup foudn at all, means: make initial backup! fake day of month ;)
                        echo -e "No Backup found for $4 with shown name $5, create initial Backup..."
                        DAYMONTH=1
                fi

                #if it is the 1st day of the month, then backup
                if [[ $DAYMONTH = "1" ]]
                then
                        if [[ $DEBUG = 1 ]]
                        then
                                echo -e "I would do now: vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0"
                        else
                                vcbMounter -a uuid:$1 -r $PATH2LUNS/$2/$6/$3_$4 -t fullvm -Q 0
                        fi
                        echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 to $PATH2LUNS/$2/$6/$3_$4 DONE\n"
                else
                        echo -e "Backup $6 of $4 with shown name $5 and\nuuid of $1 skipped, as it is day $DAYMONTH of the month. DONE\n"
                fi
                ;;

        *)
                #DEBUG
                echo -e "METHOD UNDEFINED ERROR IN do_backup for $5!"
                #DEBUG
                ;;
        esac

}

#function to remove old backups
function do_delete ()
{
        for i in $METHODS
        do
                case $i in
                ONCE)
                        #we have only one backup. keep it forever
                        ;;
                DAILY)
                        if [[ $DEBUG = 1 ]]
                        then
                                OPT=--test
                        fi
                        for j in ${BACKLUNS[*]}
                        do
                                tmpwatch $OPT --verbose $KEEPDAILY $PATH2LUNS/$j/$i
                        done

                        ;;
                WEEKLY)
                        if [[ $DEBUG = 1 ]]
                        then
                                OPT=--test
                        fi
                        for j in ${BACKLUNS[*]}
                        do
                                tmpwatch $OPT --verbose $KEEPWEEKLY $PATH2LUNS/$j/$i
                        done

                        ;;
                MONTHLY)
                        if [[ $DEBUG = 1 ]]
                        then
                                OPT=--test
                        fi
                        for j in ${BACKLUNS[*]}
                        do
                                tmpwatch $OPT --verbose $KEEPMONTHLY $PATH2LUNS/$j/$i
                        done

                        ;;
                *)
                        #DEBUG
                        echo -e "METHOD UNDEFINED ERROR IN do_delete! Aborting..."
                        exit 1
                        #DEBUG
                        ;;
                esac
        done
}

#function to find out where to backup the VM
function set_target ()
{
free=0
BACKUPDEVICE=""

for i in ${BACKLUNS[*]}
do
        temp=`vdf|grep $i|awk '{print $3}'`
        if [[ $temp -gt $free ]]
        then
                free=$temp
                BACKUPDEVICE=$i
        fi
done
}



#
#
#enough with declarations and stuff
#executing the script starts here
#
#


#create list of VMs on all machines
echo -e "We have these ESX:\n$ESX"
echo

#detect production datastores
get_prodluns

#detect backup datastores
get_backluns

#verify the found backup datastores to prevent wiping production datastores
verify_backluns

#create list with runing VMs
length=0
for i in $ESX
do
        TEMPLIST="`ssh -i $IDFILE -lroot $i vmware-cmd -l`"
        for k in $TEMPLIST
        do
                MACHINELIST[length]="$i $k"
                let length=$length+1
        done

done
let length=$length-1

intern=0
echo -e "We have these Hosts running these Machines:\n"
while [ $intern -le $length ]
do
        echo "${MACHINELIST[$intern]}"
        let intern=$intern+1

done
intern=0
echo

#delete files that have not been accessed for $HOURSOLD for each METHOD
do_delete

#parse MACHINELIST

while [ $intern -le $length ]
do

        HOST=`echo "${MACHINELIST[$intern]}"|cut -d' ' -f1`

        VMLUN="`echo "${MACHINELIST[$intern]}"|cut -d' ' -f2 |awk -F '/' '{print $4}'`"
        VMNAME="`echo "${MACHINELIST[$intern]}"|cut -d' ' -f2 |awk -F '/' '{print $5}'`"

        query=$PATH2LUNS/$VMLUN/$VMNAME/$VMNAME.vmx

        #get uuid for this machine
        VMUUID=`ssh -i $IDFILE -lroot $HOST vmware-cmd $query getconfig uuid.bios|awk -F ' = ' '{print $2}'|sed -e 's/ /-/4' -e 's/ /-/5' -e 's/ /-/7' -e 's/ //g'`

        #get shown name for this machine
        VMSHOWNNAME="`ssh -i $IDFILE -lroot $HOST vmware-cmd $query getconfig displayName |awk -F ' = ' '{print $2}'`"

        #get annotation and check if string "DO_BACKUP" is found
        VMANNO="`ssh -i $IDFILE -lroot $HOST vmware-cmd $query getconfig annotation |awk -F ' = ' '{print $2}'`"

        #is there a DO_BACKUP in the annotations? if no, skip this machine.
        #bash2 workaround for missing =~
        if [[ `echo $VMANNO|grep DO_BACKUP` ]]
        then
                echo -e "DO_BACKUP set for this machine $VMNAME, commence backup...\n"
        else
                echo -e "Skipping Machine $VMNAME, DO_BACKUP not specified in annotations...\n"
                let intern=$intern+1
                continue
        fi

        #check backup method for machine, e.g. ONCE, DAILY, WEEKLY, MONTHLY
        VMMETHOD=`echo $VMANNO|awk -F'DO_BACKUP' '{print $2}'|sed -e "s/=//g"`

        #if DO_BACKUP is not set e.g. DO_BACKUP=ONCE, then we default to DAILY backups
        if [[ "$VMMETHOD" = "" ]]
        then
                echo -e "Old style DO_BACKUP set for this machine $VMNAME, defaulting to method DAILY...\n"
                VMMETHOD=DAILY
        fi

        #get days for backup
        DAYWEEK=`date +%u`
        DAYMONTH=`date +%d`


        #always backup to Datastore with most free space
        set_target

        #error during finding lun with most free space?
        if [[ $BACKUPDEVICE = "" ]]
        then
                echo -e "ERROR: BACKUPDEVICE not set for $VMNAME skipping!!\n"
                let intern=$intern+1
                continue
        fi

        #generate Datastore list to verify existence of a VM
        for listlun in ${PRODLUNS[*]}
        do
                LUNLIST="$LUNLIST \$$listlun"
        done

        #again the bash2 workaround for esx35 compatibility
        if [[ `eval echo $LUNLIST|grep "$VMLUN"` ]]
        then
                echo -e "$VMNAME with shown name $VMSHOWNNAME and uuid $VMUUID \nlies on $VMLUN and will be backed up to $BACKUPDEVICE using method $VMMETHOD"

                do_backup $VMUUID $BACKUPDEVICE $DATE $VMNAME $VMSHOWNNAME $VMMETHOD
                if [[ $? -gt 0 ]]
                then
                        echo -e "!!ERROR!! $VMNAME/$VMSHOWNNAME/$VMUUID did NOT backup to $BACKUPDEVICE using method $VMMETHOD!!ERROR!!"
                fi
        else
                #strange things happen! we have a VM but could not find the produtcion datastore for it.
                #let us be safe, skip this machine
                echo "ERROR, NO LUN FOR $VMNAME!"
        fi

        let intern=$intern+1
done

echo -e "SCRIPT $VERSION END"

#call the helper file and create a mail that will be send
/root/sendmail.sh

exit 0

